// Copyright 2020 The Tint Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef SRC_SEM_ACCESS_CONTROL_TYPE_H_
#define SRC_SEM_ACCESS_CONTROL_TYPE_H_

#include <string>

#include "src/ast/access_control.h"
#include "src/sem/type.h"

namespace tint {
namespace sem {

/// An access control type. Holds an access setting and pointer to another type.
class AccessControl : public Castable<AccessControl, Type> {
 public:
  /// Constructor
  /// @param access the access control setting
  /// @param subtype the access controlled type
  AccessControl(ast::AccessControl::Access access, const Type* subtype);
  /// Move constructor
  AccessControl(AccessControl&&);
  ~AccessControl() override;

  /// @returns true if the access control is read only
  bool IsReadOnly() const { return access_ == ast::AccessControl::kReadOnly; }
  /// @returns true if the access control is write only
  bool IsWriteOnly() const { return access_ == ast::AccessControl::kWriteOnly; }
  /// @returns true if the access control is read/write
  bool IsReadWrite() const { return access_ == ast::AccessControl::kReadWrite; }

  /// @returns the access control value
  ast::AccessControl::Access access_control() const { return access_; }
  /// @returns the subtype type
  Type* type() const { return const_cast<Type*>(subtype_); }

  /// @returns the name for this type
  std::string type_name() const override;

  /// @param symbols the program's symbol table
  /// @returns the name for this type that closely resembles how it would be
  /// declared in WGSL.
  std::string FriendlyName(const SymbolTable& symbols) const override;

 private:
  ast::AccessControl::Access const access_;
  const Type* const subtype_;
};

}  // namespace sem
}  // namespace tint

#endif  // SRC_SEM_ACCESS_CONTROL_TYPE_H_
